import os

class Probe:

    def __init__ (self, path):
        self._probe = open(path, 'w')
        self._activity = 0
        self._dbg = 0

    def output_activity (self):
        self._probe.write(str(self._activity) + "\t" + str(self._dbg) + "\n")
        self._activity = 0
        self._dbg = 0

    def dbg (self, val):
        self._dbg = val

    def read (self, activity):
        self._activity += activity
        
    def __del__ (self):
        self._probe.close()



import sys
from .memory import Memory
from .probe  import Probe
from .exn    import *

class CPU:

    def __init__ (self, ram_size, reg_count, code, powerlog):
        self._ram = Memory(ram_size)

        self._reg = Memory(reg_count)
        self.__ra = reg_count - 1
        self.__sp = reg_count - 2

        self._code = code

        self._ip = 0 # init instruction pointer
        self._max_ip = len(self._code)
        self._reg[self.__sp] = self._ram._size # init stack pointer
        self._reg[self.__ra] = self._max_ip # init return address

        self._probe = Probe(powerlog)

    def value (self, val):
        if val[0] == 'imm':
            return val[1]
        if val[0] == 'reg':
            return self._reg[val[1]]
        if val[0] == 'mem':
            return self._ram[val[1]]
        if val[0] == 'ref':
            v = self.value(val[1])
            if len(val) == 3:
                v += self.value(val[2])
            return self._ram[v]

    def write (self, addr, val):
        if addr[0] == 'imm':
            raise WriteError
        if addr[0] == 'reg':
            self._reg[addr[1]] = val
        if addr[0] == 'mem':
            self._ram[addr[1]] = val
        if addr[0] == 'ref':
            a = self.value(addr[1])
            if len(addr) == 3:
                a += self.value(addr[2])
            self._ram[a] = val

    def execute_instruction (self, ip):
        instr = self._code[ip][0]
        opcode = instr[0]
        # opcodes with no arguments
        if opcode == 'nop':
            pass
        elif opcode == 'ret':
            return self.value(('reg', self.__ra))
        else:
            # opcode with 1 argument
            dst = instr[1]
            if opcode == 'cal':
                self.write(('reg', self.__ra), ip + 1)
                return self.value(dst)
            if opcode == 'jmp':
                return self.value(dst)
            if opcode == 'dbg':
                self._probe.dbg(dst)
            elif opcode == 'prn':
                print(self.value(dst), end='')
            elif opcode == 'prx':
                print(format(self.value(dst), '02x'), end='')
            elif opcode == 'prX':
                print(format(self.value(dst), '04x'), end='')
            elif opcode == 'prc':
                print(chr(self.value(dst)), end='')
            else:
                # opcodes with 2 arguments
                val1 = instr[2]
                if opcode == 'prs':
                    ptr = self.value(dst)
                    for i in range(0, self.value(val1)):
                        print(chr(self._ram[ptr+i]), end='')
                    print('')
                elif opcode == 'mov':
                    self.write(dst, self.value(val1))
                elif opcode == 'not':
                    self.write(dst, 0xffff ^ self.value(val1))
                else:
                    # opcodes with 3 arguments
                    val2 = instr[3]
                    if opcode == 'beq':
                        if self.value(val1) == self.value(val2):
                            return self.value(dst)
                    elif opcode == 'bne':
                        if self.value(val1) != self.value(val2):
                            return self.value(dst)
                    elif opcode == 'and':
                        self.write(dst, self.value(val1) & self.value(val2))
                    elif opcode == 'orr':
                        self.write(dst, self.value(val1) | self.value(val2))
                    elif opcode == 'xor':
                        self.write(dst, self.value(val1) ^ self.value(val2))
                    elif opcode == 'lsl':
                        self.write(dst, self.value(val1) << self.value(val2))
                    elif opcode == 'lsr':
                        self.write(dst, self.value(val1) >> self.value(val2))
                    elif opcode == 'min':
                        self.write(dst, min(self.value(val1), self.value(val2)))
                    elif opcode == 'max':
                        self.write(dst, max(self.value(val1), self.value(val2)))
                    elif opcode == 'add':
                        self.write(dst, self.value(val1) + self.value(val2))
                    elif opcode == 'sub':
                        self.write(dst, self.value(val1) - self.value(val2))
                    elif opcode == 'mul':
                        self.write(dst, self.value(val1) * self.value(val2))
                    elif opcode == 'div':
                        self.write(dst, self.value(val1) // self.value(val2))
                    elif opcode == 'mod':
                        self.write(dst, self.value(val1) % self.value(val2))
                    elif opcode == 'cmp':
                        v1 = self.value(val1)
                        v2 = self.value(val2)
                        if v1 < v2:
                            self.write(dst, 1)
                        elif v1 > v2:
                            self.write(dst, -1)
                        else:
                            self.write(dst, 0)

        return None

    def cycle (self):
        ip = self.execute_instruction(self._ip)
        if ip is not None:
            self._ip = ip
        else:
            self._ip += 1
        self._probe.read(self._ram.get_activity())
        self._probe.read(self._reg.get_activity())
        self._probe.output_activity()
        sys.stdout.flush()

        if self._ip < 0 or self._ip >= self._max_ip:
            return False
        return True

    def run (self):
        try:
            while self.cycle():
                pass

        except AddrError as e:
            print('Invalid address ' + str(e.addr) +
                  ' on line ' + str(self._code[self._ip][1][1]) +
                  ' of file ' + self._code[self._ip][1][0],
                  file=sys.stderr)

        except WriteError:
            print('Invalid write ' +
                  ' on line ' + str(self._code[self._ip][1][1]) +
                  ' of file ' + self._code[self._ip][1][0],
                  file=sys.stderr)



import sys
import os
from .exn import *

class Compiler:

    def __init__ (self):
        self._ln = []
        self._files = []
        self._count = 0
        self._labels = {}
        self._code = []

    def arit (self, i, a):
        if len(i) != a + 1:
            raise ParseError('arity')

    def addr (self, a, noref = False):
        if a[0] == 'r':
            if a[1:].isdigit():
                return ('reg', int(a[1:]))
            else:
                raise ParseError('register')
        if a[0] == '@':
            if a[1:].isdigit():
                return ('mem', int(a[1:]))
            else:
                raise ParseError('memory')
        if a[0] == '!' and not noref:
            if ',' in a:
                r = a[1:].split(',', 1)
                return ('ref', self.val(r[0], True), self.val(r[1], True))
            else:
                return ('ref', self.val(a[1:], True))
        if noref:
            raise ParseError('reference')
        raise ParseError('address')

    def val (self, v, noref = False):
        if v[0] == '#':
            try:
                return ('imm', int(v[1:]))
            except:
                raise ParseError('immediate')
        else:
            return self.addr(v, noref)

    def lbl (self, l):
        if l.isidentifier():
            return ('lbl', l)
        else:
            return self.val(l)

    def instr (self, line):
        line = line.split(';', 1)[0].strip()
        if line == '':
            return None

        if ':' in line:
            line = line.split(':', 1)
            lbl = line[0].strip()
            if lbl in self._labels:
                raise DuplicateLabel(lbl)
            else:
                self._labels[lbl] = self._count, self._files[-1], self._ln[-1]
            line = line[1]

        line = line.split(';', 1)[0].strip()
        if line == '':
            return None

        i = line.split()

        if i[0] == 'nop':
            self.arit(i, 0)
            return [i[0]]

        if i[0] == 'ret':
            self.arit(i, 0)
            return [i[0]]

        if i[0] == 'cal':
            self.arit(i, 1)
            return [i[0], self.lbl(i[1])]

        if i[0] == 'jmp':
            self.arit(i, 1)
            return [i[0], self.lbl(i[1])]

        if i[0] == 'dbg':
            self.arit(i, 1)
            return [i[0], i[1]]

        if i[0] == 'prn':
            self.arit(i, 1)
            return [i[0], self.val(i[1])]

        if i[0] == 'prx':
            self.arit(i, 1)
            return [i[0], self.val(i[1])]

        if i[0] == 'prX':
            self.arit(i, 1)
            return [i[0], self.val(i[1])]

        if i[0] == 'prc':
            self.arit(i, 1)
            return [i[0], self.val(i[1])]

        if i[0] == 'prs':
            self.arit(i, 2)
            return [i[0], self.val(i[1]), self.val(i[2])]

        if i[0] == 'mov':
            self.arit(i, 2)
            return [i[0], self.addr(i[1]), self.val(i[2])]

        if i[0] == 'not':
            self.arit(i, 2)
            return [i[0], self.addr(i[1]), self.val(i[2])]

        if i[0] == 'beq':
            self.arit(i, 3)
            return [i[0], self.lbl(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'bne':
            self.arit(i, 3)
            return [i[0], self.lbl(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'and':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'orr':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'xor':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'lsl':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'lsr':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'min':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'max':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'add':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'sub':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'mul':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'div':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'mod':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        if i[0] == 'cmp':
            self.arit(i, 3)
            return [i[0], self.addr(i[1]), self.val(i[2]), self.val(i[3])]

        raise ParseError('opcode')

    def compile_file (self, path):
        inp = open(path, 'r')
        self._files.append(os.path.basename(path))
        self._ln.append(1)
        for l in inp:
            if '.include' in l:
                self.compile_file(
                    os.path.join(os.path.dirname(path),
                                 l.split('.include', 1)[1].strip()))
            elif '.use' in l:
                seselab = os.path.dirname(sys.modules[__package__].__file__)
                self.compile_file(
                    os.path.join(seselab,
                                 l.split('.use', 1)[1].strip() + '.asm'))
            else:
                a = self.instr(l)
                if a is not None:
                    self._code.append([a, [self._files[-1], self._ln[-1]]])
                    self._count += 1
            self._ln[-1] += 1
        self._files.pop()
        self._ln.pop()
        inp.close()

    def compile (self, path):
        self._code.append([['jmp', ('lbl', 'main')], ['_', -1]])
        self._count += 1
        self.compile_file(os.path.abspath(path))
        for instr in self._code:
            if instr[0][0] in ('cal', 'jmp', 'beq', 'bne'):
                if instr[0][1][0] == 'lbl':
                    lbl = instr[0][1][1]
                    if lbl in self._labels:
                        instr[0][1] = 'imm', self._labels[lbl][0]
                        instr[1].append(lbl)
                    else:
                        raise LabelNotFound(lbl, instr[1][0], instr[1][1])
        for lbl in self._labels:
            instr = self._code[self._labels[lbl][0]]
            if len(instr[1]) == 2:
                instr[1].append(None)
            instr[1].append(lbl)
        return self._code

    def run (self, path):
        try:
            return self.compile(path)

        except FileNotFoundError as e:
            print('File not found: ' + e.filename)

        except ParseError as e:
            print('Parse error: ' + e.err +
                  ' in ' + self._files[-1] +
                  ' on line ' + str(self._ln[-1]),
                  file=sys.stderr)

        except DuplicateLabel as e:
            print('Duplicate label: ' + e.lbl +
                  ' in ' + self._files[-1] +
                  ' on line ' + str(self._ln[-1]) +
                  ' (first declared in ' + self._labels[e.lbl][1] +
                  ' on line ' + str(self._labels[e.lbl][2]) + ')',
                  file=sys.stderr)

        except LabelNotFound as e:
            print('Label not found: ' + e.lbl +
                  ' (called in ' + e.file +
                  ' on line ' + str(e.line) + ')',
                  file=sys.stderr)

        sys.exit(1)



class ParseError (Exception):
    def __init__ (self, err):
        self.err = err

class DuplicateLabel (Exception):
    def __init__ (self, lbl):
        self.lbl = lbl

class LabelNotFound (Exception):
    def __init__ (self, lbl, file, line):
        self.lbl = lbl
        self.file = file
        self.line = line

class AddrError (Exception):
    def __init__ (self, addr):
        self.addr = addr

class WriteError (Exception):
    def __init__ (self):
        pass



import sys
import signal
from random import randint
from .compiler import Compiler
from .cpu import CPU

def usage ():
    print('Usage: seselab [-i] <code.asm> [power.log]')
    print(' * -i:        enable fault injection during execution if present')
    print(' * code.asm:  file with seselab assembly code to execute')
    print(' * power.log: file where to output simulated power consumption '
          + '(default to /dev/null)')
    sys.exit(1)

def main ():
    if len(sys.argv) < 2:
        usage()
    fault = sys.argv[1] == '-i'
    if fault:
        sys.argv.pop(1)
        if len(sys.argv) < 2:
            usage()
    asmcode = sys.argv[1]
    powerlog = '/dev/null' if len(sys.argv) < 3 else sys.argv[2]
    program = Compiler().run(asmcode)
    cpu = CPU(1048576, 32, program, powerlog)

    def inject (sig, frame):
        where = input('\n> Inject fault? (r = fault register, s = skip instruction, q = quit) ')
        if where == 'q':
            sys.exit(0)
        elif where == 's':
            skip = int(input('> How many instruction to skip? '))
            cpu._ip += skip
            print('! ' + str(skip) + ' instructions skipped')
        elif where == 'r':
            reg = int(input('> Which register? (0-'+str(cpu._reg._size)+') '))
            what = input('> Zero or random? (z = zero, r = random) ')
            if what == 'z':
                cpu._reg._mem[reg] = 0
                print('! r' + str(reg) + ' zeroized')
            elif what == 'r':
                cpu._reg._mem[reg] = randint(0, cpu._ram._size)
                print('! r' + str(reg) + ' randomized')
        print('> Resuming…')

    if fault:
        signal.signal(signal.SIGINT, inject)
    else:
        signal.signal(signal.SIGINT, lambda s, f: sys.exit(0))

    cpu.run()



from setuptools import setup, find_packages

def readme ():
    with open('README.md') as me:
        return me.read()

setup(name='seselab',
      version='0.1.1',
      description='SeseLab: a software platform for teaching physical attacks',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://pablo.rauzy.name/software.html#seselab',
      author='Pablo Rauzy',
      author_email='pr_NOSPAM'+chr(64)+'up8.edu',
      license='GNU AGPL v3+',
      packages=find_packages(),
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'seselab = seselab.bench:main'
          ]
      })



from .exn import *

def hw (n):
    return format(n, 'b').count('1')

class Memory:

    def __init__ (self, size):
        self._size = size
        self._mem = [0 for x in range(0, size)]
        self._addr = 0 # address bus
        self._hw = 0 # total Hamming weight
        self._hd = 0 # current cycle Hamming distance

    def __setitem__ (self, addr, val):
        if addr < 0 or addr >= self._size:
            raise AddrError(addr)

        self._hw -= hw(self._mem[addr])
        self._hw += hw(val)

        self._hd += hw(self._mem[addr] ^ val)

        self._addr = addr

        self._mem[addr] = val

    def __getitem__ (self, addr):
        if addr < 0 or addr >= self._size:
            raise AddrError(addr)

        self._addr = addr

        return self._mem[addr]

    def get_activity (self):
        activity = (self._hw // 2) + self._hd + hw(self._addr)
        self._hd = 0
        return activity
